@@ -1,5 +1,6 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE BangPatterns        #-}
+{-# LANGUAGE CPP                 #-}
+{-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 {-|
@@ -17,8 +18,12 @@ module Snap.Http.Server.Config
   , emptyConfig
   , defaultConfig
   , commandLineConfig
+  , extendedCommandLineConfig
   , completeConfig
 
+  , optDescrs
+  , fmapOpt
+
   , getAccessLog
   , getBackend
   , getBind
@@ -30,6 +35,7 @@ module Snap.Http.Server.Config
   , getLocale
   , getOther
   , getPort
+  , getProxyType
   , getSSLBind
   , getSSLCert
   , getSSLKey
@@ -47,6 +53,7 @@ module Snap.Http.Server.Config
   , setLocale
   , setOther
   , setPort
+  , setProxyType
   , setSSLBind
   , setSSLCert
   , setSSLKey
@@ -54,7 +61,7 @@ module Snap.Http.Server.Config
   , setVerbose
   ) where
 
-
+------------------------------------------------------------------------------
 import           Blaze.ByteString.Builder
 import           Blaze.ByteString.Builder.Char8
 import           Control.Exception (SomeException)
@@ -73,6 +80,7 @@ import           Prelude hiding (catch)
 import           Snap.Core
 import           Snap.Iteratee ((>==>), enumBuilder)
 import           Snap.Internal.Debug (debug)
+import           Snap.Util.Proxy
 import           System.Console.GetOpt
 import           System.Environment hiding (getEnv)
 #ifndef PORTABLE
@@ -80,7 +88,6 @@ import           System.Posix.Env
 #endif
 import           System.Exit
 import           System.IO
-
 ------------------------------------------------------------------------------
 import           Snap.Internal.Http.Server (requestErrorMessage)
 
@@ -103,9 +110,9 @@ data ConfigLog = ConfigNoLog                        -- ^ no logging
                | ConfigIoLog (ByteString -> IO ())  -- ^ log custom IO handler
 
 instance Show ConfigLog where
-    show ConfigNoLog       = "ConfigNoLog"
-    show (ConfigFileLog f) = "ConfigFileLog " ++ show f
-    show (ConfigIoLog _)   = "ConfigIoLog"
+    show ConfigNoLog       = "no log"
+    show (ConfigFileLog f) = "log to file " ++ show f
+    show (ConfigIoLog _)   = "custom logging handler"
 
 ------------------------------------------------------------------------------
 -- | A record type which represents partial configurations (for 'httpServe')
@@ -133,6 +140,7 @@ data Config m a = Config
     , defaultTimeout :: Maybe Int
     , other          :: Maybe a
     , backend        :: Maybe ConfigBackend
+    , proxyType      :: Maybe ProxyType
     }
 
 instance Show (Config m a) where
@@ -151,6 +159,7 @@ instance Show (Config m a) where
                      , "verbose: "        ++ _verbose
                      , "defaultTimeout: " ++ _defaultTimeout
                      , "backend: "        ++ _backend
+                     , "proxyType: "      ++ _proxyType
                      ]
 
       where
@@ -168,6 +177,7 @@ instance Show (Config m a) where
         _verbose        = show $ verbose        c
         _defaultTimeout = show $ defaultTimeout c
         _backend        = show $ backend        c
+        _proxyType      = show $ proxyType      c
 
 
 ------------------------------------------------------------------------------
@@ -196,28 +206,30 @@ instance Monoid (Config m a) where
         , defaultTimeout = Nothing
         , other          = Nothing
         , backend        = Nothing
+        , proxyType      = Nothing
         }
 
     a `mappend` b = Config
-        { hostname       = ov hostname       a b
-        , accessLog      = ov accessLog      a b
-        , errorLog       = ov errorLog       a b
-        , locale         = ov locale         a b
-        , port           = ov port           a b
-        , bind           = ov bind           a b
-        , sslport        = ov sslport        a b
-        , sslbind        = ov sslbind        a b
-        , sslcert        = ov sslcert        a b
-        , sslkey         = ov sslkey         a b
-        , compression    = ov compression    a b
-        , verbose        = ov verbose        a b
-        , errorHandler   = ov errorHandler   a b
-        , defaultTimeout = ov defaultTimeout a b
-        , other          = ov other          a b
-        , backend        = ov backend        a b
+        { hostname       = ov hostname
+        , accessLog      = ov accessLog
+        , errorLog       = ov errorLog
+        , locale         = ov locale
+        , port           = ov port
+        , bind           = ov bind
+        , sslport        = ov sslport
+        , sslbind        = ov sslbind
+        , sslcert        = ov sslcert
+        , sslkey         = ov sslkey
+        , compression    = ov compression
+        , verbose        = ov verbose
+        , errorHandler   = ov errorHandler
+        , defaultTimeout = ov defaultTimeout
+        , other          = ov other
+        , backend        = ov backend
+        , proxyType      = ov proxyType
         }
       where
-        ov f x y = getLast $! (mappend `on` (Last . f)) x y
+        ov f = getLast $! (mappend `on` (Last . f)) a b
 
 
 ------------------------------------------------------------------------------
@@ -251,7 +263,9 @@ defaultConfig = mempty
 
 
 ------------------------------------------------------------------------------
--- | The hostname of the HTTP server
+-- | The hostname of the HTTP server. This field has the same format as an HTTP
+-- @Host@ header; if a @Host@ header came in with the request, we use that,
+-- otherwise we default to this value specified in the configuration.
 getHostname       :: Config m a -> Maybe ByteString
 getHostname = hostname
 
@@ -263,7 +277,10 @@ getAccessLog = accessLog
 getErrorLog       :: Config m a -> Maybe ConfigLog
 getErrorLog = errorLog
 
--- | The locale to use
+-- | Gets the locale to use. Locales are used on Unix only, to set the
+-- @LANG@\/@LC_ALL@\/etc. environment variable. For instance if you set the
+-- locale to \"@en_US@\", we'll set the relevant environment variables to
+-- \"@en_US.UTF-8@\".
 getLocale         :: Config m a -> Maybe String
 getLocale = locale
 
@@ -312,6 +329,9 @@ getOther = other
 getBackend :: Config m a -> Maybe ConfigBackend
 getBackend = backend
 
+getProxyType :: Config m a -> Maybe ProxyType
+getProxyType = proxyType
+
 
 ------------------------------------------------------------------------------
 setHostname       :: ByteString              -> Config m a -> Config m a
@@ -362,6 +382,9 @@ setOther x c = c { other = Just x }
 setBackend        :: ConfigBackend           -> Config m a -> Config m a
 setBackend x c = c { backend = Just x }
 
+setProxyType      :: ProxyType               -> Config m a -> Config m a
+setProxyType x c = c { proxyType = Just x }
+
 
 ------------------------------------------------------------------------------
 completeConfig :: (MonadSnap m) => Config m a -> IO (Config m a)
@@ -369,7 +392,7 @@ completeConfig config = do
     when noPort $ hPutStrLn stderr
         "no port specified, defaulting to port 8000"
 
-    return $ cfg `mappend` cfg'
+    return $! cfg `mappend` cfg'
 
   where
     cfg = defaultConfig `mappend` config
@@ -396,10 +419,12 @@ toString = T.unpack . T.decodeUtf8
 
 
 ------------------------------------------------------------------------------
-options :: MonadSnap m =>
-           Config m a
-        -> [OptDescr (Maybe (Config m a))]
-options defaults =
+-- | Returns a description of the snap command line options suitable for use
+-- with "System.Console.GetOpt".
+optDescrs :: MonadSnap m =>
+             Config m a         -- ^ the configuration defaults.
+          -> [OptDescr (Maybe (Config m a))]
+optDescrs defaults =
     [ Option [] ["hostname"]
              (ReqArg (Just . setConfig setHostname . bsFromString) "NAME")
              $ "local hostname" ++ defaultC getHostname
@@ -437,30 +462,45 @@ options defaults =
              $ "don't have an error log"
     , Option ['c'] ["compression"]
              (NoArg $ Just $ setConfig setCompression True)
-             $ "use gzip compression on responses"
+             $ "use gzip compression on responses" ++
+               defaultB getCompression "compressed" "uncompressed"
     , Option ['t'] ["timeout"]
              (ReqArg (\t -> Just $ mempty {
                               defaultTimeout = Just $ read t
                             }) "SECS")
-             $ "set default timeout in seconds"
+             $ "set default timeout in seconds" ++ defaultC defaultTimeout
     , Option [] ["no-compression"]
              (NoArg $ Just $ setConfig setCompression False)
-             $ "serve responses uncompressed"
+             $ "serve responses uncompressed" ++
+               defaultB compression "compressed" "uncompressed"
     , Option ['v'] ["verbose"]
              (NoArg $ Just $ setConfig setVerbose True)
-             $ "print server status updates to stderr"
+             $ "print server status updates to stderr" ++
+               defaultC getVerbose
     , Option ['q'] ["quiet"]
              (NoArg $ Just $ setConfig setVerbose False)
-             $ "do not print anything to stderr"
+             $ "do not print anything to stderr" ++
+               defaultB getVerbose "verbose" "quiet"
+    , Option [] ["proxy"]
+             (ReqArg (\t -> Just $ setConfig setProxyType $ read t)
+                     "X_Forwarded_For")
+             $ concat [ "Set --proxy=X_Forwarded_For if your snap application "
+                      , "is behind an HTTP reverse proxy to ensure that "
+                      , "rqRemoteAddr is set properly."
+                      , defaultC getProxyType ]
     , Option ['h'] ["help"]
              (NoArg Nothing)
              $ "display this help and exit"
     ]
+
   where
-    setConfig f c = f c mempty
-    conf          = defaultConfig `mappend` defaults
-    defaultC f    = maybe "" ((", default " ++) . show) $ f conf
-    defaultO f    = maybe ", default off" ((", default " ++) . show) $ f conf
+    setConfig f c  = f c mempty
+    conf           = defaultConfig `mappend` defaults
+    defaultB f y n = maybe "" (\b -> ", default " ++ if b
+                                                       then y
+                                                       else n) $ f conf
+    defaultC f     = maybe "" ((", default " ++) . show) $ f conf
+    defaultO f     = maybe ", default off" ((", default " ++) . show) $ f conf
 
 
 ------------------------------------------------------------------------------
@@ -478,6 +518,7 @@ defaultErrorHandler e = do
                . modifyResponseBody
                      (>==> enumBuilder (fromByteString msg))
                $ emptyResponse
+
   where
     smsg req = toByteString $ requestErrorMessage req e
 
@@ -490,28 +531,56 @@ defaultErrorHandler e = do
 
 
 ------------------------------------------------------------------------------
--- | Returns a 'Config' obtained from parsing the options specified on the
--- command-line.
+-- | Returns a 'Config' obtained from parsing command-line options, using the
+-- default Snap 'OptDescr' set.
 --
 -- On Unix systems, the locale is read from the @LANG@ environment variable.
 commandLineConfig :: MonadSnap m
                   => Config m a
-                  -- ^ default configuration. This is combined with
-                  -- 'defaultConfig' to obtain default values to use if the
-                  -- given parameter is specified on the command line. Usually
-                  -- it is fine to use 'emptyConfig' here.
+                      -- ^ default configuration. This is combined with
+                      -- 'defaultConfig' to obtain default values to use if the
+                      -- given parameter is specified on the command line.
+                      -- Usually it is fine to use 'emptyConfig' here.
                   -> IO (Config m a)
-commandLineConfig defaults = do
+commandLineConfig defaults = extendedCommandLineConfig (optDescrs defaults) f defaults
+  where
+    -- Here getOpt can ever change the "other" field, because we only use the
+    -- Snap OptDescr list. The combining function will never be invoked.
+    f = undefined
+
+
+------------------------------------------------------------------------------
+-- | Returns a 'Config' obtained from parsing command-line options, using the
+-- default Snap 'OptDescr' set as well as a list of user OptDescrs. User
+-- OptDescrs use the \"other\" field (accessible using 'getOther' and
+-- 'setOther') to store additional command-line option state. These are
+-- combined using a user-defined combining function.
+--
+-- On Unix systems, the locale is read from the @LANG@ environment variable.
+
+extendedCommandLineConfig :: MonadSnap m
+                          => [OptDescr (Maybe (Config m a))]
+                             -- ^ User options.
+                          -> (a -> a -> a)
+                             -- ^ State for multiple invoked user command-line
+                             -- options will be combined using this function.
+                          -> Config m a
+                             -- ^ default configuration. This is combined with
+                             -- Snap's 'defaultConfig' to obtain default values
+                             -- to use if the given parameter is specified on
+                             -- the command line. Usually it is fine to use
+                             -- 'emptyConfig' here.
+                          -> IO (Config m a)
+extendedCommandLineConfig opts combiningFunction defaults = do
     args <- getArgs
     prog <- getProgName
 
-    let opts = options defaults
-
-    result <- either (usage prog opts)
+    result <- either (usage prog)
                      return
                      (case getOpt Permute opts args of
                         (f, _, []  ) -> maybe (Left []) Right $
-                                        fmap mconcat $ sequence f
+                                        fmap (foldl' combine mempty) $
+                                        sequence f
                         (_, _, errs) -> Left errs)
 
 #ifndef PORTABLE
@@ -524,7 +593,7 @@ commandLineConfig defaults = do
 #endif
 
   where
-    usage prog opts errs = do
+    usage prog errs = do
         let hdr = "Usage:\n  " ++ prog ++ " [OPTION...]\n\nOptions:"
         let msg = concat errs ++ usageInfo hdr opts
         hPutStrLn stderr msg
@@ -532,3 +601,25 @@ commandLineConfig defaults = do
 #ifndef PORTABLE
     upToUtf8 = takeWhile $ \c -> isAlpha c || '_' == c
 #endif
+
+    combine !a !b = a `mappend` b `mappend` newOther
+      where
+        -- combined is only a Just if both a and b have other fields, and then
+        -- we use the combining function. Config's mappend picks the last
+        -- "Just" in the other list.
+        combined = do
+            x <- getOther a
+            y <- getOther b
+            return $! combiningFunction x y
+
+        newOther = mempty { other = combined }
+
+fmapArg :: (a -> b) -> ArgDescr a -> ArgDescr b
+fmapArg f (NoArg a) = NoArg (f a)
+fmapArg f (ReqArg g s) = ReqArg (f . g) s
+fmapArg f (OptArg g s) = OptArg (f . g) s
+
+fmapOpt :: (a -> b) -> OptDescr a -> OptDescr b
+fmapOpt f (Option s l d e) = Option s l (fmapArg f d) e
+
+
