@@ -29,11 +29,12 @@ import qualified Data.ByteString.Char8 as SC
 import qualified Data.ByteString.Lazy as L
 import           Data.ByteString.Internal (c2w, w2c)
 import qualified Data.ByteString.Nums.Careless.Int as Cvt
+import           Data.Enumerator.Internal
 import           Data.Int
 import           Data.IORef
 import           Data.List (foldl')
 import qualified Data.Map as Map
-import           Data.Maybe (catMaybes, fromJust, fromMaybe)
+import           Data.Maybe (catMaybes, fromJust, fromMaybe, isJust)
 import           Data.Monoid
 import           Data.Time
 import           Data.Typeable
@@ -46,9 +47,9 @@ import           System.Posix.Types (FileOffset)
 import           System.Locale
 ------------------------------------------------------------------------------
 import           System.FastLogger (timestampedLogEntry, combinedLogEntry)
-import           Snap.Core (EscapeHttpException (..))
 import           Snap.Internal.Http.Types
 import           Snap.Internal.Debug
+import           Snap.Internal.Exceptions (EscapeHttpException (..))
 import           Snap.Internal.Http.Parser
 import           Snap.Internal.Http.Server.Date
 
@@ -77,7 +78,7 @@ import qualified Paths_snap_server as V
 -- Note that we won't be bothering end users with this -- the details will be
 -- hidden inside the Snap monad
 type ServerHandler = (ByteString -> IO ())
-                  -> (Int -> IO ())
+                  -> ((Int -> Int) -> IO ())
                   -> Request
                   -> Iteratee ByteString IO (Request,Response)
 
@@ -95,8 +96,9 @@ data ListenPort =
 
 ------------------------------------------------------------------------------
 instance Show ListenPort where
-    show (HttpPort b p) =
-        concat [ "http://", SC.unpack b, ":", show p, "/" ]
+    show (HttpPort  b p    ) =
+        concat [ "http://" , SC.unpack b, ":", show p, "/" ]
+
     show (HttpsPort b p _ _) =
         concat [ "https://", SC.unpack b, ":", show p, "/" ]
 
@@ -198,7 +200,7 @@ httpServe defaultTimeout ports mevType localHostname alog' elog'
             logE elog "Server.httpServe: BACKEND STOPPED"
 
     --------------------------------------------------------------------------
-    bindPort (HttpPort  baddr port)          = bindHttp  baddr port
+    bindPort (HttpPort  baddr port         ) = bindHttp  baddr port
     bindPort (HttpsPort baddr port cert key) =
         TLS.bindHttps baddr port cert key
 
@@ -263,7 +265,7 @@ runHTTP :: Int                           -- ^ default timeout
         -> Iteratee ByteString IO ()     -- ^ write end of socket
         -> (FilePath -> Int64 -> Int64 -> IO ())
                                          -- ^ sendfile end
-        -> (Int -> IO ())                -- ^ timeout tickler
+        -> ((Int -> Int) -> IO ())       -- ^ timeout tickler
         -> IO ()
 runHTTP defaultTimeout alog elog handler lh sinfo readEnd writeEnd onSendFile
         tickle =
@@ -284,7 +286,8 @@ runHTTP defaultTimeout alog elog handler lh sinfo readEnd writeEnd onSendFile
                      , fromShow $ remoteAddress sinfo
                      , fromByteString "]: "
                      , fromByteString "an exception escaped to toplevel:\n"
-                     , fromShow e ]
+                     , fromShow e
+                     ]
 
     go = do
         buf <- allocBuffer 16384
@@ -312,7 +315,8 @@ requestErrorMessage req e =
             , fromByteString "\nrequest:\n"
             , fromShow $ show req
             , fromByteString "\n"
-            , msgB ]
+            , msgB
+            ]
   where
     msgB = mconcat [
              fromByteString "A web handler threw an exception. Details:\n"
@@ -347,7 +351,7 @@ httpSession :: Int
             -> Buffer                        -- ^ builder buffer
             -> (FilePath -> Int64 -> Int64 -> IO ())
                                              -- ^ sendfile continuation
-            -> (Int -> IO ())                -- ^ timeout tickler
+            -> ((Int -> Int) -> IO ())       -- ^ timeout modifier
             -> ServerHandler                 -- ^ handler procedure
             -> ServerMonad ()
 httpSession defaultTimeout writeEnd' buffer onSendFile tickle handler = do
@@ -359,7 +363,7 @@ httpSession defaultTimeout writeEnd' buffer onSendFile tickle handler = do
     debug "Server.httpSession: receiveRequest finished"
 
     -- successfully got a request, so restart timer
-    liftIO $ tickle defaultTimeout
+    liftIO $ tickle (max defaultTimeout)
 
     case mreq of
       (Just req) -> do
@@ -409,8 +413,11 @@ httpSession defaultTimeout writeEnd' buffer onSendFile tickle handler = do
                            "sending response"
 
           date <- liftIO getDateString
+          let insHeader = H.set "Server" sERVER_HEADER
           let ins = H.set "Date" date .
-                    H.set "Server" sERVER_HEADER
+                    if isJust (getHeader "Server" rsp)
+                       then id
+                       else insHeader
           let rsp' = updateHeaders ins rsp
           (bytesSent,_) <- sendResponse req rsp' buffer writeEnd onSendFile
                            `catch` errCatch "sending response" req
@@ -446,7 +453,8 @@ httpSession defaultTimeout writeEnd' buffer onSendFile tickle handler = do
           mconcat [ fromByteString "httpSession caught an exception during "
                   , fromByteString phase
                   , fromByteString " phase:\n"
-                  , requestErrorMessage req e ]
+                  , requestErrorMessage req e
+                  ]
         throw ExceptionAlreadyCaught
 
 
@@ -468,7 +476,8 @@ checkExpect100Continue req writeEnd = do
                          , fromShow major
                          , fromWord8 $ c2w '.'
                          , fromShow minor
-                         , fromByteString " 100 Continue\r\n\r\n" ]
+                         , fromByteString " 100 Continue\r\n\r\n"
+                         ]
         liftIO $ runIteratee
                    ((enumBS (toByteString hl) >==> enumEOF) $$ writeEnd)
         return ()
@@ -490,7 +499,8 @@ return411 req writeEnd = do
                          , fromWord8 $ c2w '.'
                          , fromShow minor
                          , fromByteString " 411 Length Required\r\n\r\n"
-                         , fromByteString "411 Length Required\r\n" ]
+                         , fromByteString "411 Length Required\r\n"
+                         ]
         liftIO $ runIteratee
                    ((enumBS (toByteString hl) >==> enumEOF) $$ writeEnd)
         return ()
@@ -501,7 +511,8 @@ receiveRequest :: Iteratee ByteString IO () -> ServerMonad (Maybe Request)
 receiveRequest writeEnd = do
     debug "receiveRequest: entered"
     mreq <- {-# SCC "receiveRequest/parseRequest" #-} lift $
-            iterateeDebugWrapper "parseRequest" parseRequest
+            iterateeDebugWrapper "parseRequest" $
+            joinI' $ takeNoMoreThan maxHeadersSize $$ parseRequest
     debug "receiveRequest: parseRequest returned"
 
     case mreq of
@@ -510,12 +521,16 @@ receiveRequest writeEnd = do
           setEnumerator req'
           req  <- parseForm req'
           checkConnectionClose (rqVersion req) (rqHeaders req)
-          return $ Just req
+          return $! Just req
 
       Nothing     -> return Nothing
 
-
   where
+    --------------------------------------------------------------------------
+    -- TODO(gdc): make this a policy decision (expose in
+    -- Snap.Http.Server.Config)
+    maxHeadersSize = 256 * 1024
+
     --------------------------------------------------------------------------
     -- check: did the client specify "transfer-encoding: chunked"? then we
     -- have to honor that.
@@ -613,8 +628,10 @@ receiveRequest writeEnd = do
                 e st'
 
             liftIO $ writeIORef (rqBody req) $ SomeEnumerator e'
-            return $ req { rqParams = Map.unionWith (++) (rqParams req)
-                                        newParams }
+            return $! req { rqParams = Map.unionWith (++) (rqParams req)
+                                         newParams
+                          , rqPostParams = newParams
+                          }
 
 
     --------------------------------------------------------------------------
@@ -635,30 +652,29 @@ receiveRequest writeEnd = do
             -- will override in "setEnumerator"
             enum <- liftIO $ newIORef $ SomeEnumerator (enumBS "")
 
-            return $ Request serverName
-                             serverPort
-                             remoteAddr
-                             rport
-                             localAddr
-                             lport
-                             localHostname
-                             secure
-                             hdrs
-                             enum
-                             mbContentLength
-                             method
-                             version
-                             cookies
-                             snapletPath
-                             pathInfo
-                             contextPath
-                             uri
-                             queryString
-                             params
+            return $! Request serverName
+                              serverPort
+                              remoteAddr
+                              rport
+                              localAddr
+                              lport
+                              localHostname
+                              secure
+                              hdrs
+                              enum
+                              mbContentLength
+                              method
+                              version
+                              cookies
+                              pathInfo
+                              contextPath
+                              uri
+                              queryString
+                              params
+                              params
+                              Map.empty
 
       where
-        snapletPath = ""        -- TODO: snaplets in v0.2
-
         dropLeadingSlash s = maybe s f mbS
           where
             f (a,s') = if a == c2w '/' then s' else s
@@ -741,9 +757,14 @@ sendResponse req rsp' buffer writeEnd' onSendFile = do
         outstep <- lift $ runIteratee $
                    iterateeDebugWrapper "countBytes writeEnd" $
                    countBytes writeEnd
+
+        let bufferFunc = if getBufferingMode rsp
+                           then unsafeBuilderToByteString (return buffer)
+                           else I.map (toByteString . (`mappend` flush))
+
         (x,bs) <- mapIter fromByteString toByteString
-                          (enum $$ joinI $ unsafeBuilderToByteString
-                              (return buffer) outstep)
+                          (enum $$ joinI $ bufferFunc outstep)
+
         debug $ "sendResponse: whenEnum: " ++ show bs ++
                 " bytes enumerated"
 
@@ -858,7 +879,7 @@ sendResponse req rsp' buffer writeEnd' onSendFile = do
         {-# SCC "setFileSize" #-}
         do
             fs <- liftM fromIntegral $ liftIO $ getFileSize fp
-            return $ r { rspContentLength = Just fs }
+            return $! r { rspContentLength = Just fs }
 
 
     --------------------------------------------------------------------------
@@ -892,7 +913,7 @@ sendResponse req rsp' buffer writeEnd' onSendFile = do
             z <- case rspBody r'' of
                    (Enum _)                  -> return r''
                    (SendFile f Nothing)      -> setFileSize f r''
-                   (SendFile _ (Just (s,e))) -> return $
+                   (SendFile _ (Just (s,e))) -> return $!
                                                 setContentLength (e-s) r''
 
             case rspContentLength z of
