@@ -238,7 +238,7 @@ acceptCallback defaultTimeout back handler
   where
     go = runSession defaultTimeout back handler sock
     cleanup = [ Handler $ \(_ :: TimeoutException) -> return ()
-              , Handler $ \(e :: AsyncException)   -> return ()
+              , Handler $ \(_ :: AsyncException)   -> return ()
               , Handler $ \(e :: SomeException) ->
                   elog $ S.concat [ "libev.acceptCallback: "
                                   , S.pack . map c2w $ show e ]
@@ -525,7 +525,7 @@ runSession defaultTimeout backend handler lsock fd = do
                                       (enumerate conn session)
                                       (writeOut defaultTimeout conn session)
                                       (sendFile defaultTimeout conn session)
-                                      (setTimeout conn)
+                                      (modifyTimeout conn)
                 )
 
 
@@ -570,24 +570,40 @@ instance Exception TimeoutException
 
 
 ------------------------------------------------------------------------------
-tickleTimeout :: Connection -> Int -> IO ()
-tickleTimeout conn tm = do
-    debug "Libev.tickleTimeout"
-    now  <- getCurrentDateTime
-    prev <- readIORef ref
-    let !n = max (now + toEnum tm) prev
-    writeIORef ref n
+modifyTimeout :: Connection -> (Int -> Int) -> IO ()
+modifyTimeout conn f = do
+    debug "Libev.modifyTimeout"
+    !prev <- readIORef tt
+    !now  <- getCurrentDateTime
+
+    let !remaining    = fromEnum $ max 0 (prev - now)
+    let !newRemaining = f remaining
+    let !newTimeout   = now + toEnum newRemaining
+
+    writeIORef tt $! now + toEnum newRemaining
+
+    -- Here the question is: do we reset the ev timer? If we're extending the
+    -- timeout, the ev manual suggests it's more efficient to let the timer
+    -- lapse and re-arm. If we're shortening the timeout, we need to update the
+    -- timer so it fires when it's supposed to.
+    when (newTimeout < prev) $ withMVar loopLock $ \_ -> do
+        evTimerSetRepeat tmr $! toEnum newRemaining
+        evTimerAgain loop tmr
+        -- wake up the event loop so it can be apprised of the changes
+        evAsyncSend loop asyncObj
 
   where
-    ref = _timerTimeoutTime conn
+    tt       = _timerTimeoutTime conn
+    backend  = _backend conn
+    asyncObj = _asyncObj backend
+    loopLock = _loopLock backend
+    loop     = _evLoop backend
+    tmr      = _timerObj conn
 
 
 ------------------------------------------------------------------------------
-setTimeout :: Connection -> Int -> IO ()
-setTimeout conn tm = do
-    debug "Libev.tickleTimeout"
-    now       <- getCurrentDateTime
-    writeIORef (_timerTimeoutTime conn) (now + toEnum tm)
+tickleTimeout :: Connection -> Int -> IO ()
+tickleTimeout conn = modifyTimeout conn . max
 
 
 ------------------------------------------------------------------------------
